from ctypes import c_int32
from enum import Enum
from oead import (
    S32,
    F32,
    FixedSafeString32,
    FixedSafeString64,
    FixedSafeString256,
    Vector2f,
    Vector3f,
    Vector4f,
)
from oead.byml import Hash, Array
from typing import List, Tuple
from zlib import crc32
import re

from . import overrides


class BFUFlag:
    def __init__(self, flag: Hash = None) -> None:
        self.data_name = ""
        self.delete_rev = -1
        self.is_event_associated = False
        self.is_one_trigger = False
        self.is_program_readable = True
        self.is_program_writable = True
        self.is_save = False
        self.reset_type = 0
        if flag:
            if not BFUFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.data_name = flag["DataName"]
            self.delete_rev = flag["DeleteRev"].v
            self.is_event_associated = flag["IsEventAssociated"]
            self.is_one_trigger = flag["IsOneTrigger"]
            self.is_program_readable = flag["IsProgramReadable"]
            self.is_program_writable = flag["IsProgramWritable"]
            self.is_save = flag["IsSave"]
            self.reset_type = flag["ResetType"].v

    def __eq__(self, other):
        if self is other:
            return True
        if (
            not self.data_name == other.data_name
            or not self.delete_rev == other.delete_rev
            or not self.hash_value == other.hash_value
            or not self.is_event_associated == other.is_event_associated
            or not self.is_one_trigger == other.is_one_trigger
            or not self.is_program_readable == other.is_program_readable
            or not self.is_program_writable == other.is_program_writable
            or not self.is_save == other.is_save
            or not self.reset_type == other.reset_type
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["DeleteRev"]) == S32
            assert type(hash["IsEventAssociated"]) == bool
            assert type(hash["IsOneTrigger"]) == bool
            assert type(hash["IsProgramReadable"]) == bool
            assert type(hash["IsProgramWritable"]) == bool
            assert type(hash["IsSave"]) == bool
            assert type(hash["ResetType"]) == S32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        """Converts flag to an oead.Hash meant for a gamedata file"""
        r = Hash()
        r["DataName"] = self.data_name
        r["DeleteRev"] = S32(self.delete_rev)
        r["HashValue"] = S32(self.hash_value)
        r["IsEventAssociated"] = self.is_event_associated
        r["IsOneTrigger"] = self.is_one_trigger
        r["IsProgramReadable"] = self.is_program_readable
        r["IsProgramWritable"] = self.is_program_writable
        r["IsSave"] = self.is_save
        r["ResetType"] = S32(self.reset_type)
        return r

    def to_sv_Hash(self) -> Hash:
        """Converts flag to an oead.Hash meant for a saveformat_#.bgsvdata"""
        r = Hash()
        r["DataName"] = self.data_name
        r["HashValue"] = S32(self.hash_value)
        return r

    def exists(self) -> bool:
        """Returns True if the flag contains any data"""
        return not self.hash_value == 0

    def name_contains(self, name: str) -> bool:
        """Returns true if the flag's name contains the passed string"""
        return name in self.data_name

    @property
    def data_name(self) -> str:
        return self._data_name

    @data_name.setter
    def data_name(self, name: str) -> None:
        self._data_name = name
        self._hash_value = c_int32(crc32(name.encode("utf-8"))).value

    @property
    def delete_rev(self) -> int:
        return self._delete_rev

    @delete_rev.setter
    def delete_rev(self, drev: int) -> None:
        self._delete_rev = drev

    @property
    def hash_value(self) -> int:
        return self._hash_value

    @property
    def is_event_associated(self) -> bool:
        return self._is_event_associated

    @is_event_associated.setter
    def is_event_associated(self, event_assoc: bool) -> None:
        self._is_event_associated = event_assoc

    @property
    def is_one_trigger(self) -> bool:
        return self._is_one_trigger

    @is_one_trigger.setter
    def is_one_trigger(self, is_one: bool) -> None:
        self._is_one_trigger = is_one

    @property
    def is_program_readable(self) -> bool:
        return self._is_program_readable

    @is_program_readable.setter
    def is_program_readable(self, is_read: bool) -> None:
        self._is_program_readable = is_read

    @property
    def is_program_writable(self) -> bool:
        return self._is_program_writable

    @is_program_writable.setter
    def is_program_writable(self, is_write: bool) -> None:
        self._is_program_writable = is_write

    @property
    def is_save(self) -> bool:
        return self._is_save

    @is_save.setter
    def is_save(self, save: bool) -> None:
        self._is_save = save

    @property
    def reset_type(self) -> int:
        return self._reset_type

    @reset_type.setter
    def reset_type(self, reset_type: int) -> None:
        self._reset_type = reset_type

    @property
    def is_revival(self) -> bool:
        return False

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        OVERRIDES = overrides["STANDARD_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_IS_EVENT_ASSOCIATED"].items():
            if re.search(regex, self.data_name):
                self.is_event_associated = value
        for regex, value in OVERRIDES["OVERRIDE_IS_ONE_TRIGGER"].items():
            if re.search(regex, self.data_name):
                self.is_one_trigger = value
        for regex, value in OVERRIDES["OVERRIDE_IS_PROGRAM_READABLE"].items():
            if re.search(regex, self.data_name):
                self.is_program_readable = value
        for regex, value in OVERRIDES["OVERRIDE_IS_PROGRAM_WRITABLE"].items():
            if re.search(regex, self.data_name):
                self.is_program_writable = value
        for regex, value in OVERRIDES["OVERRIDE_IS_SAVE"].items():
            if re.search(regex, self.data_name):
                self.is_save = value
        for regex, value in OVERRIDES["OVERRIDE_RESET_TYPE"].items():
            if re.search(regex, self.data_name):
                self.reset_type = value


class BoolFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(BoolFlag, self).__init__(flag=flag)
        self.category = -1
        self.init_value = 0
        self.max_value = True
        self.min_value = False
        self.is_revival = kwargs["revival"] if "revival" in kwargs else False
        if flag:
            if not BoolFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.category = flag["Category"].v if "Category" in flag else -1
            self.init_value = flag["InitValue"].v
            self.max_value = flag["MaxValue"]
            self.min_value = flag["MinValue"]

    def __eq__(self, other):
        if not type(other) == BoolFlag:
            return NotImplemented
        if not super(BoolFlag, self).__eq__(super(BoolFlag, other)):
            return False
        if (
            not self.category == other.category
            or not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            if "Category" in hash:
                assert type(hash["Category"]) == S32
            assert type(hash["InitValue"]) == S32
            assert type(hash["MaxValue"]) == bool
            assert type(hash["MinValue"]) == bool
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(BoolFlag, self).to_Hash()
        if not self.category == -1:
            r["Category"] = S32(self.category)
        r["InitValue"] = S32(self.init_value)
        r["MaxValue"] = self.max_value
        r["MinValue"] = self.min_value
        return r

    @property
    def category(self) -> int:
        return self._category

    @category.setter
    def category(self, value: int) -> None:
        self._category = value

    @property
    def init_value(self) -> int:
        return self._init_value

    @init_value.setter
    def init_value(self, value: bool) -> None:
        self._init_value = value

    @property
    def max_value(self) -> bool:
        return self._max_value

    @max_value.setter
    def max_value(self, value: bool) -> None:
        self._max_value = value

    @property
    def min_value(self) -> bool:
        return self._min_value

    @min_value.setter
    def min_value(self, value: bool) -> None:
        self._min_value = value

    @property
    def is_revival(self) -> bool:
        return self._revival

    @is_revival.setter
    def is_revival(self, revival: bool) -> None:
        self._revival = revival

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(BoolFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["BOOL_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_BOOL_CATEGORY"].items():
            if re.search(regex, self.data_name):
                self.category = value
        for regex, value in OVERRIDES["OVERRIDE_BOOL_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value


class BoolArrayFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(BoolArrayFlag, self).__init__(flag=flag)
        self.init_value = [0]
        self.max_value = True
        self.min_value = False
        if flag:
            if not BoolArrayFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = [value.v for value in flag["InitValue"][0]["Values"]]
            self.max_value = flag["MaxValue"]
            self.min_value = flag["MinValue"]

    def __eq__(self, other):
        if not type(other) == BoolArrayFlag:
            return NotImplemented
        if not super(BoolArrayFlag, self).__eq__(super(BoolArrayFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == Array
            assert type(hash["InitValue"][0]) == Hash
            assert "Values" in hash["InitValue"][0]
            for value in hash["InitValue"][0]["Values"]:
                assert type(value) == S32
            assert type(hash["MaxValue"]) == bool
            assert type(hash["MinValue"]) == bool
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(BoolArrayFlag, self).to_Hash()
        array_one = Array()
        array_one.append(Hash())
        array_two = Array([S32(value) for value in self.init_value])
        array_one[0]["Values"] = array_two
        r["InitValue"] = array_one
        r["MaxValue"] = self.max_value
        r["MinValue"] = self.min_value
        return r

    @property
    def init_value(self) -> List[int]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: List[int]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> bool:
        return self._max_value

    @max_value.setter
    def max_value(self, value: bool) -> None:
        self._max_value = value

    @property
    def min_value(self) -> bool:
        return self._min_value

    @min_value.setter
    def min_value(self, value: bool) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(BoolArrayFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["BOOL_ARRAY_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_BOOL_ARRAY_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_BOOL_ARRAY_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_BOOL_ARRAY_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class S32Flag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(S32Flag, self).__init__(flag=flag)
        self.init_value = 0
        self.max_value = 2147483647
        self.min_value = 0
        self.is_revival = kwargs["revival"] if "revival" in kwargs else False
        if flag:
            if not S32Flag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self._init_value = flag["InitValue"].v
            self._max_value = flag["MaxValue"].v
            self._min_value = flag["MinValue"].v

    def __eq__(self, other):
        if not type(other) == S32Flag:
            return NotImplemented
        if not super(S32Flag, self).__eq__(super(S32Flag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == S32
            assert type(hash["MaxValue"]) == S32
            assert type(hash["MinValue"]) == S32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(S32Flag, self).to_Hash()
        r["InitValue"] = S32(self._init_value)
        r["MaxValue"] = S32(self._max_value)
        r["MinValue"] = S32(self._min_value)
        return r

    @property
    def init_value(self) -> int:
        return self._init_value

    @init_value.setter
    def init_value(self, value: int) -> None:
        self._init_value = value

    @property
    def max_value(self) -> int:
        return self._max_value

    @max_value.setter
    def max_value(self, value: int) -> None:
        self._max_value = value

    @property
    def min_value(self) -> int:
        return self._min_value

    @min_value.setter
    def min_value(self, value: int) -> None:
        self._min_value = value

    @property
    def is_revival(self) -> bool:
        return self._revival

    @is_revival.setter
    def is_revival(self, revival: bool) -> None:
        self._revival = revival

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(S32Flag, self).use_name_to_override_params()
        OVERRIDES = overrides["S32_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_S32_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_S32_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_S32_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class S32ArrayFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(S32ArrayFlag, self).__init__(flag=flag)
        self.init_value = [0]
        self.max_value = 6553500
        self.min_value = -1
        if flag:
            if not S32ArrayFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = [int(value) for value in flag["InitValue"][0]["Values"]]
            self.max_value = flag["MaxValue"].v
            self.min_value = flag["MinValue"].v

    def __eq__(self, other):
        if not type(other) == S32ArrayFlag:
            return NotImplemented
        if not super(S32ArrayFlag, self).__eq__(super(S32ArrayFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == Array
            assert type(hash["InitValue"][0]) == Hash
            assert "Values" in hash["InitValue"][0]
            for value in hash["InitValue"][0]["Values"]:
                assert type(value) == S32
            assert type(hash["MaxValue"]) == S32
            assert type(hash["MinValue"]) == S32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(S32ArrayFlag, self).to_Hash()
        array_one = Array()
        array_one.append(Hash())
        array_two = Array([S32(value) for value in self.init_value])
        array_one[0]["Values"] = array_two
        r["InitValue"] = array_one
        r["MaxValue"] = S32(self.max_value)
        r["MinValue"] = S32(self.min_value)
        return r

    @property
    def init_value(self) -> List[int]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: List[int]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> int:
        return self._max_value

    @max_value.setter
    def max_value(self, value: int) -> None:
        self._max_value = value

    @property
    def min_value(self) -> int:
        return self._min_value

    @min_value.setter
    def min_value(self, value: int) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(S32ArrayFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["S32_ARRAY_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_S32_ARRAY_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_S32_ARRAY_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_S32_ARRAY_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class F32Flag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(F32Flag, self).__init__(flag=flag)
        self.init_value = 0.0
        self.max_value = 1000000.0
        self.min_value = 0.0
        if flag:
            if not F32Flag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = flag["InitValue"].v
            self.max_value = flag["MaxValue"].v
            self.min_value = flag["MinValue"].v

    def __eq__(self, other):
        if not type(other) == F32Flag:
            return NotImplemented
        if not super(F32Flag, self).__eq__(super(F32Flag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == F32
            assert type(hash["MaxValue"]) == F32
            assert type(hash["MinValue"]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(F32Flag, self).to_Hash()
        r["InitValue"] = F32(self.init_value)
        r["MaxValue"] = F32(self.max_value)
        r["MinValue"] = F32(self.min_value)
        return r

    @property
    def init_value(self) -> float:
        return self._init_value

    @init_value.setter
    def init_value(self, value: float) -> None:
        self._init_value = value

    @property
    def max_value(self) -> float:
        return self._max_value

    @max_value.setter
    def max_value(self, value: float) -> None:
        self._max_value = value

    @property
    def min_value(self) -> float:
        return self._min_value

    @min_value.setter
    def min_value(self, value: float) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(F32Flag, self).use_name_to_override_params()
        OVERRIDES = overrides["F32_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_F32_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_F32_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_F32_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class F32ArrayFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(F32ArrayFlag, self).__init__(flag=flag)
        self.init_value = [0.0]
        self.max_value = 360.0
        self.min_value = -1.0
        if flag:
            if not F32ArrayFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = [value.v for value in flag["InitValue"][0]["Values"]]
            self.max_value = flag["MaxValue"].v
            self.min_value = flag["MinValue"].v

    def __eq__(self, other):
        if not type(other) == F32ArrayFlag:
            return NotImplemented
        if not super(F32ArrayFlag, self).__eq__(super(F32ArrayFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == Array
            assert type(hash["InitValue"][0]) == Hash
            assert "Values" in hash["InitValue"][0]
            for value in hash["InitValue"][0]["Values"]:
                assert type(value) == F32
            assert type(hash["MaxValue"]) == F32
            assert type(hash["MinValue"]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(F32ArrayFlag, self).to_Hash()
        array_one = Array()
        array_one.append(Hash())
        array_two = Array([F32(value) for value in self.init_value])
        array_one[0]["Values"] = array_two
        r["InitValue"] = array_one
        r["MaxValue"] = F32(self.max_value)
        r["MinValue"] = F32(self.min_value)
        return r

    @property
    def init_value(self) -> List[float]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: List[float]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> float:
        return self._max_value

    @max_value.setter
    def max_value(self, value: float) -> None:
        self._max_value = value

    @property
    def min_value(self) -> float:
        return self._min_value

    @min_value.setter
    def min_value(self, value: float) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(F32ArrayFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["F32_ARRAY_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_F32_ARRAY_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_F32_ARRAY_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_F32_ARRAY_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class StringFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(StringFlag, self).__init__(flag=flag)
        self.init_value = ""
        self.max_value = ""
        self.min_value = ""
        if flag:
            if not StringFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = flag["InitValue"]
            self.max_value = flag["MaxValue"]
            self.min_value = flag["MinValue"]

    def __eq__(self, other):
        if not isinstance(other, StringFlag):
            return NotImplemented
        if not super(StringFlag, self).__eq__(super(StringFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == str
            assert type(hash["MaxValue"]) == str
            assert type(hash["MinValue"]) == str
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(StringFlag, self).to_Hash()
        return r

    @property
    def init_value(self) -> str:
        return self._init_value

    @init_value.setter
    def init_value(self, value: str) -> None:
        self._init_value = value

    @property
    def max_value(self) -> str:
        return self._max_value

    @max_value.setter
    def max_value(self, value: str) -> None:
        self._max_value = value

    @property
    def min_value(self) -> str:
        return self._min_value

    @min_value.setter
    def min_value(self, value: str) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(StringFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["STRING_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_STRING_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_STRING_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_STRING_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class String32Flag(StringFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(String32Flag, self).__init__(flag=flag)

    def to_Hash(self) -> Hash:
        r = super(String32Flag, self).to_Hash()
        # r["InitValue"] = FixedSafeString32(self.init_value)
        r["InitValue"] = self.init_value
        # r["MaxValue"] = FixedSafeString32(self.max_value)
        r["MaxValue"] = self.max_value
        # r["MinValue"] = FixedSafeString32(self.min_value)
        r["MinValue"] = self.min_value
        return r


class String64Flag(StringFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(String64Flag, self).__init__(flag=flag)

    def to_Hash(self) -> Hash:
        r = super(String64Flag, self).to_Hash()
        # r["InitValue"] = FixedSafeString64(self.init_value)
        r["InitValue"] = self.init_value
        # r["MaxValue"] = FixedSafeString64(self.max_value)
        r["MaxValue"] = self.max_value
        # r["MinValue"] = FixedSafeString64(self.min_value)
        r["MinValue"] = self.min_value
        return r


class String256Flag(StringFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(String256Flag, self).__init__(flag=flag)

    def to_Hash(self) -> Hash:
        r = super(String256Flag, self).to_Hash()
        # r["InitValue"] = FixedSafeString256(self.init_value)
        r["InitValue"] = self.init_value
        # r["MaxValue"] = FixedSafeString256(self.max_value)
        r["MaxValue"] = self.max_value
        # r["MinValue"] = FixedSafeString256(self.min_value)
        r["MinValue"] = self.min_value
        return r


class StringArrayFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(StringArrayFlag, self).__init__(flag=flag)
        self.init_value = [""]
        self.max_value = ""
        self.min_value = ""
        if flag:
            if not StringArrayFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = [value for value in flag["InitValue"][0]["Values"]]
            self.max_value = flag["MaxValue"]
            self.min_value = flag["MinValue"]

    def __eq__(self, other):
        if not isinstance(other, StringArrayFlag):
            return NotImplemented
        if not super(StringArrayFlag, self).__eq__(super(StringArrayFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == Array
            assert type(hash["InitValue"][0]) == Hash
            assert "Values" in hash["InitValue"][0]
            for value in hash["InitValue"][0]["Values"]:
                assert type(value) == str
            assert type(hash["MaxValue"]) == str
            assert type(hash["MinValue"]) == str
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(StringArrayFlag, self).to_Hash()
        return r

    @property
    def init_value(self) -> List[str]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: List[str]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> str:
        return self._max_value

    @max_value.setter
    def max_value(self, value: str) -> None:
        self._max_value = value

    @property
    def min_value(self) -> str:
        return self._min_value

    @min_value.setter
    def min_value(self, value: str) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(StringArrayFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["STRING_ARRAY_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_STRING_ARRAY_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_STRING_ARRAY_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_STRING_ARRAY_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class String64ArrayFlag(StringArrayFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(String64ArrayFlag, self).__init__(flag=flag)

    def to_Hash(self) -> Hash:
        r = super(String64ArrayFlag, self).to_Hash()
        array_one = Array()
        array_one.append(Hash())
        # array_two = [FixedSafeString64(value) for value in self.init_value]
        array_two = Array(self.init_value)
        array_one[0]["Values"] = array_two
        r["InitValue"] = array_one
        # r["MaxValue"] = FixedSafeString64(self.max_value)
        r["MaxValue"] = self.max_value
        # r["MinValue"] = FixedSafeString64(self.min_value)
        r["MinValue"] = self.min_value
        return r


class String256ArrayFlag(StringArrayFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(String256ArrayFlag, self).__init__(flag=flag)

    def to_Hash(self) -> Hash:
        r = super(String256ArrayFlag, self).to_Hash()
        array_one = Array()
        array_one.append(Hash())
        # array_two = [FixedSafeString256(value) for value in self.init_value]
        array_two = Array(self.init_value)
        array_one[0]["Values"] = array_two
        r["InitValue"] = array_one
        # r["MaxValue"] = FixedSafeString256(self.max_value)
        r["MaxValue"] = self.max_value
        # r["MinValue"] = FixedSafeString256(self.min_value)
        r["MinValue"] = self.min_value
        return r


class Vec2Flag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(Vec2Flag, self).__init__(flag=flag)
        self.init_value = (0.0, 0.0)
        self.max_value = (255.0, 255.0)
        self.min_value = (0.0, 0.0)
        if flag:
            if not Vec2Flag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = (flag["InitValue"][0][0].v, flag["InitValue"][0][1].v)
            self.max_value = (flag["MaxValue"][0][0].v, flag["MaxValue"][0][1].v)
            self.min_value = (flag["MinValue"][0][0].v, flag["MinValue"][0][1].v)

    def __eq__(self, other):
        if not type(other) == Vec2Flag:
            return NotImplemented
        if not super(Vec2Flag, self).__eq__(super(Vec2Flag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"][0][0]) == F32
            assert type(hash["InitValue"][0][1]) == F32
            assert type(hash["MaxValue"][0][0]) == F32
            assert type(hash["MaxValue"][0][1]) == F32
            assert type(hash["MinValue"][0][0]) == F32
            assert type(hash["MinValue"][0][1]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(Vec2Flag, self).to_Hash()
        array = Array()
        vec = Array()
        vec.append(F32(self.init_value[0]))
        vec.append(F32(self.init_value[1]))
        array.append(vec)
        r["InitValue"] = array
        array = Array()
        vec = Array()
        vec.append(F32(self.max_value[0]))
        vec.append(F32(self.max_value[1]))
        array.append(vec)
        r["MaxValue"] = array
        array = Array()
        vec = Array()
        vec.append(F32(self.min_value[0]))
        vec.append(F32(self.min_value[1]))
        array.append(vec)
        r["MinValue"] = array
        return r

    @property
    def init_value(self) -> Tuple[float, float]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: Tuple[float, float]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> Tuple[float, float]:
        return self._max_value

    @max_value.setter
    def max_value(self, value: Tuple[float, float]) -> None:
        self._max_value = value

    @property
    def min_value(self) -> Tuple[float, float]:
        return self._min_value

    @min_value.setter
    def min_value(self, value: Tuple[float, float]) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(Vec2Flag, self).use_name_to_override_params()
        OVERRIDES = overrides["VEC2_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_VEC2_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC2_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC2_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class Vec2ArrayFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(Vec2ArrayFlag, self).__init__(flag=flag)
        self.init_value = [(0.0, 0.0)]
        self.max_value = (255.0, 255.0)
        self.min_value = (0.0, 0.0)
        if flag:
            if not Vec2ArrayFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            vec_array = []
            for vector in flag["InitValue"][0]["Values"]:
                vec_array.append((vector[0][0].v, vector[0][1].v))
            self.init_value = vec_array
            self.max_value = (flag["MaxValue"][0][0].v, flag["MaxValue"][0][1].v)
            self.min_value = (flag["MinValue"][0][0].v, flag["MinValue"][0][1].v)

    def __eq__(self, other):
        if not type(other) == Vec2ArrayFlag:
            return NotImplemented
        if not super(Vec2ArrayFlag, self).__eq__(super(Vec2ArrayFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == Array
            assert type(hash["InitValue"][0]) == Hash
            assert "Values" in hash["InitValue"][0]
            for value in hash["InitValue"][0]["Values"]:
                assert type(value[0][0]) == F32
                assert type(value[0][1]) == F32
            assert type(hash["MaxValue"][0][0]) == F32
            assert type(hash["MaxValue"][0][1]) == F32
            assert type(hash["MinValue"][0][0]) == F32
            assert type(hash["MinValue"][0][1]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(Vec2ArrayFlag, self).to_Hash()
        vec_array = Array()
        vec_array.append(Hash())
        vec_array[0]["Values"] = Array()
        for i in range(len(self.init_value)):
            vector = self.init_value[i]
            vec = Array()
            vec.append(F32(vector[0]))
            vec.append(F32(vector[1]))
            vec_array[0]["Values"].append(Array())
            vec_array[0]["Values"][i].append(vec)
        r["InitValue"] = vec_array
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.max_value[0]))
        vec.append(F32(self.max_value[1]))
        vec_array.append(vec)
        r["MaxValue"] = vec_array
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.min_value[0]))
        vec.append(F32(self.min_value[1]))
        vec_array.append(vec)
        r["MinValue"] = vec_array
        return r

    @property
    def init_value(self) -> List[Tuple[float, float]]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: List[Tuple[float, float]]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> Tuple[float, float]:
        return self._max_value

    @max_value.setter
    def max_value(self, value: Tuple[float, float]) -> None:
        self._max_value = value

    @property
    def min_value(self) -> Tuple[float, float]:
        return self._min_value

    @min_value.setter
    def min_value(self, value: Tuple[float, float]) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(Vec2ArrayFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["VEC2_ARRAY_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_VEC2_ARRAY_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC2_ARRAY_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC2_ARRAY_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class Vec3Flag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(Vec3Flag, self).__init__(flag=flag)
        self.init_value = (0.0, 0.0, 0.0)
        self.max_value = (100000.0, 100000.0, 100000.0)
        self.min_value = (-100000.0, -100000.0, -100000.0)
        if flag:
            if not Vec3Flag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = (
                flag["InitValue"][0][0].v,
                flag["InitValue"][0][1].v,
                flag["InitValue"][0][2].v,
            )
            self.max_value = (
                flag["MaxValue"][0][0].v,
                flag["MaxValue"][0][1].v,
                flag["MaxValue"][0][2].v,
            )
            self.min_value = (
                flag["MinValue"][0][0].v,
                flag["MinValue"][0][1].v,
                flag["MinValue"][0][2].v,
            )

    def __eq__(self, other):
        if not type(other) == Vec3Flag:
            return NotImplemented
        if not super(Vec3Flag, self).__eq__(super(Vec3Flag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"][0][0]) == F32
            assert type(hash["InitValue"][0][1]) == F32
            assert type(hash["InitValue"][0][2]) == F32
            assert type(hash["MaxValue"][0][0]) == F32
            assert type(hash["MaxValue"][0][1]) == F32
            assert type(hash["MaxValue"][0][2]) == F32
            assert type(hash["MinValue"][0][0]) == F32
            assert type(hash["MinValue"][0][1]) == F32
            assert type(hash["MinValue"][0][2]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(Vec3Flag, self).to_Hash()
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.init_value[0]))
        vec.append(F32(self.init_value[1]))
        vec.append(F32(self.init_value[2]))
        vec_array.append(vec)
        r["InitValue"] = vec_array
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.max_value[0]))
        vec.append(F32(self.max_value[1]))
        vec.append(F32(self.max_value[2]))
        vec_array.append(vec)
        r["MaxValue"] = vec_array
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.min_value[0]))
        vec.append(F32(self.min_value[1]))
        vec.append(F32(self.min_value[2]))
        vec_array.append(vec)
        r["MinValue"] = vec_array
        return r

    @property
    def init_value(self) -> Tuple[float, float, float]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: Tuple[float, float, float]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> Tuple[float, float, float]:
        return self._max_value

    @max_value.setter
    def max_value(self, value: Tuple[float, float, float]) -> None:
        self._max_value = value

    @property
    def min_value(self) -> Tuple[float, float, float]:
        return self._min_value

    @min_value.setter
    def min_value(self, value: Tuple[float, float, float]) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(Vec3Flag, self).use_name_to_override_params()
        OVERRIDES = overrides["VEC3_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_VEC3_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC3_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC3_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class Vec3ArrayFlag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(Vec3ArrayFlag, self).__init__(flag=flag)
        self.init_value = [(0.0, 0.0, 0.0)]
        self.max_value = (255.0, 255.0, 255.0)
        self.min_value = (0.0, 0.0, 0.0)
        if flag:
            if not Vec3ArrayFlag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            vec_array = []
            for vector in flag["InitValue"][0]["Values"]:
                vec_array.append((vector[0][0].v, vector[0][1].v, vector[0][2].v))
            self.init_value = vec_array
            self.max_value = (
                flag["MaxValue"][0][0].v,
                flag["MaxValue"][0][1].v,
                flag["MaxValue"][0][2].v,
            )
            self.min_value = (
                flag["MinValue"][0][0].v,
                flag["MinValue"][0][1].v,
                flag["MinValue"][0][2].v,
            )

    def __eq__(self, other):
        if not type(other) == Vec3ArrayFlag:
            return NotImplemented
        if not super(Vec3ArrayFlag, self).__eq__(super(Vec3ArrayFlag, other)):
            return False
        if (
            not self.init_value == other.init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"]) == Array
            assert type(hash["InitValue"][0]) == Hash
            assert "Values" in hash["InitValue"][0]
            for value in hash["InitValue"][0]["Values"]:
                assert type(value[0][0]) == F32
                assert type(value[0][1]) == F32
                assert type(value[0][2]) == F32
            assert type(hash["MaxValue"][0][0]) == F32
            assert type(hash["MaxValue"][0][1]) == F32
            assert type(hash["MaxValue"][0][2]) == F32
            assert type(hash["MinValue"][0][0]) == F32
            assert type(hash["MinValue"][0][1]) == F32
            assert type(hash["MinValue"][0][2]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(Vec3ArrayFlag, self).to_Hash()
        vec_array = Array()
        vec_array.append(Hash())
        vec_array[0]["Values"] = Array()
        for i in range(len(self.init_value)):
            vector = self.init_value[i]
            vec = Array()
            vec.append(F32(vector[0]))
            vec.append(F32(vector[1]))
            vec.append(F32(vector[2]))
            vec_array[0]["Values"].append(Array())
            vec_array[0]["Values"][i].append(vec)
        r["InitValue"] = vec_array
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.max_value[0]))
        vec.append(F32(self.max_value[1]))
        vec.append(F32(self.max_value[2]))
        vec_array.append(vec)
        r["MaxValue"] = vec_array
        vec_array = Array()
        vec = Array()
        vec.append(F32(self.min_value[0]))
        vec.append(F32(self.min_value[1]))
        vec.append(F32(self.min_value[2]))
        vec_array.append(vec)
        r["MinValue"] = vec_array
        return r

    @property
    def init_value(self) -> List[Tuple[float, float, float]]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: List[Tuple[float, float, float]]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> Tuple[float, float, float]:
        return self._max_value

    @max_value.setter
    def max_value(self, value: Tuple[float, float, float]) -> None:
        self._max_value = value

    @property
    def min_value(self) -> Tuple[float, float, float]:
        return self._min_value

    @min_value.setter
    def min_value(self, value: Tuple[float, float, float]) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(Vec3ArrayFlag, self).use_name_to_override_params()
        OVERRIDES = overrides["VEC3_ARRAY_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_VEC3_ARRAY_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC3_ARRAY_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC3_ARRAY_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value


class Vec4Flag(BFUFlag):
    def __init__(self, flag: Hash = None, **kwargs) -> None:
        super(Vec4Flag, self).__init__(flag=flag)
        self.init_value = (0.0, 0.0, 0.0, 0.0)
        self.max_value = (255.0, 255.0, 255.0, 255.0)
        self.min_value = (0.0, 0.0, 0.0, 0.0)
        if flag:
            if not Vec4Flag.validate_Hash(flag):
                raise AttributeError(f"{flag['DataName']} is malformed.")
            self.init_value = (
                flag["InitValue"][0][0].v,
                flag["InitValue"][0][1].v,
                flag["InitValue"][0][2].v,
                flag["InitValue"][0][3].v,
            )
            self.max_value = (
                flag["MaxValue"][0][0].v,
                flag["MaxValue"][0][1].v,
                flag["MaxValue"][0][2].v,
                flag["MaxValue"][0][3].v,
            )
            self.min_value = (
                flag["MinValue"][0][0].v,
                flag["MinValue"][0][1].v,
                flag["MinValue"][0][2].v,
                flag["MinValue"][0][3].v,
            )

    def __eq__(self, other):
        if not type(other) == Vec4Flag:
            return NotImplemented
        if not super(Vec4Flag, self).__eq__(super(Vec4Flag, other)):
            return False
        if (
            not self.init_value == other._init_value
            or not self.max_value == other.max_value
            or not self.min_value == other.min_value
        ):
            return False
        return True

    @staticmethod
    def validate_Hash(hash: Hash) -> bool:
        """Returns True if all Hash properties are valid"""
        try:
            assert type(hash["InitValue"][0][0]) == F32
            assert type(hash["InitValue"][0][1]) == F32
            assert type(hash["InitValue"][0][2]) == F32
            assert type(hash["InitValue"][0][3]) == F32
            assert type(hash["MaxValue"][0][0]) == F32
            assert type(hash["MaxValue"][0][1]) == F32
            assert type(hash["MaxValue"][0][2]) == F32
            assert type(hash["MaxValue"][0][3]) == F32
            assert type(hash["MinValue"][0][0]) == F32
            assert type(hash["MinValue"][0][1]) == F32
            assert type(hash["MinValue"][0][2]) == F32
            assert type(hash["MinValue"][0][3]) == F32
            return True
        except AssertionError:
            return False

    def to_Hash(self) -> Hash:
        r = super(Vec4Flag, self).to_Hash()
        array = Array()
        vec = Array()
        vec.append(F32(self.init_value[0]))
        vec.append(F32(self.init_value[1]))
        vec.append(F32(self.init_value[2]))
        vec.append(F32(self.init_value[3]))
        array.append(vec)
        r["InitValue"] = array
        array = Array()
        vec = Array()
        vec.append(F32(self.max_value[0]))
        vec.append(F32(self.max_value[1]))
        vec.append(F32(self.max_value[2]))
        vec.append(F32(self.max_value[3]))
        array.append(vec)
        r["MaxValue"] = array
        array = Array()
        vec = Array()
        vec.append(F32(self.min_value[0]))
        vec.append(F32(self.min_value[1]))
        vec.append(F32(self.min_value[2]))
        vec.append(F32(self.min_value[3]))
        array.append(vec)
        r["MinValue"] = array
        return r

    @property
    def init_value(self) -> Tuple[float, float, float, float]:
        return self._init_value

    @init_value.setter
    def init_value(self, value: Tuple[float, float, float, float]) -> None:
        self._init_value = value

    @property
    def max_value(self) -> Tuple[float, float, float, float]:
        return self._max_value

    @max_value.setter
    def max_value(self, value: Tuple[float, float, float, float]) -> None:
        self._max_value = value

    @property
    def min_value(self) -> Tuple[float, float, float, float]:
        return self._min_value

    @min_value.setter
    def min_value(self, value: Tuple[float, float, float, float]) -> None:
        self._min_value = value

    def use_name_to_override_params(self) -> None:
        """
        Sets flag parameters to those mandated by the overrides.
        In general, this function should always be called right
        before finalizing the flag creation/changes to ensure
        that certain values that should always be the same for
        certain flag types are upheld.
        """
        super(Vec4Flag, self).use_name_to_override_params()
        OVERRIDES = overrides["VEC4_OVERRIDES"]
        for regex, value in OVERRIDES["OVERRIDE_VEC4_INIT_VALUE"].items():
            if re.search(regex, self.data_name):
                self.init_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC4_MAX_VALUE"].items():
            if re.search(regex, self.data_name):
                self.max_value = value
        for regex, value in OVERRIDES["OVERRIDE_VEC4_MIN_VALUE"].items():
            if re.search(regex, self.data_name):
                self.min_value = value



from oead.byml import Array, Hash
from typing import Dict, List, Set

from . import BGDATA_MAPPING
from .flag import (
    BFUFlag,
    BoolFlag,
    BoolArrayFlag,
    S32Flag,
    S32ArrayFlag,
    F32Flag,
    F32ArrayFlag,
    String32Flag,
    String64Flag,
    String64ArrayFlag,
    String256Flag,
    String256ArrayFlag,
    Vec2Flag,
    Vec2ArrayFlag,
    Vec3Flag,
    Vec3ArrayFlag,
    Vec4Flag,
)


FLAG_MAPPING = {
    "bool_data": BoolFlag,
    "bool_array_data": BoolArrayFlag,
    "s32_data": S32Flag,
    "s32_array_data": S32ArrayFlag,
    "f32_data": F32Flag,
    "f32_array_data": F32ArrayFlag,
    "string_data": String32Flag,
    "string64_data": String64Flag,
    "string64_array_data": String64ArrayFlag,
    "string256_data": String256Flag,
    "string256_array_data": String256ArrayFlag,
    "vector2f_data": Vec2Flag,
    "vector2f_array_data": Vec2ArrayFlag,
    "vector3f_data": Vec3Flag,
    "vector3f_array_data": Vec3ArrayFlag,
    "vector4f_data": Vec4Flag,
}
IGNORED_SAVE_FLAGS = [
    "AlbumPictureIndex",
    "IsGet_Obj_AmiiboItem",
    "CaptionPictSize",
    "SeakSensorPictureIndex",
    "AoC_HardMode_Enabled",
    "FamouseValue",
    "SaveDistrictName",
    "LastSaveTime_Lower",
    "GameClear",
    "IsChangedByDebug",
    "SaveLocationName",
    "IsSaveByAuto",
    "LastSaveTime_Upper",
    "IsLogicalDelete",
    "GyroOnOff",
    "PlayReport_CtrlMode_Ext",
    "PlayReport_CtrlMode_Free",
    "NexUniqueID_Upper",
    "MiniMapDirection",
    "CameraRLReverse",
    "JumpButtonChange",
    "TextRubyOnOff",
    "VoiceLanguage",
    "PlayReport_CtrlMode_Console_Free",
    "PlayReport_PlayTime_Handheld",
    "BalloonTextOnOff",
    "PlayReport_AudioChannel_Other",
    "PlayReport_AudioChannel_5_1ch",
    "NexIsPosTrackUploadAvailableCache",
    "NexsSaveDataUploadIntervalHoursCache",
    "NexUniqueID_Lower",
    "TrackBlockFileNumber",
    "Option_LatestAoCVerPlayed",
    "NexPosTrackUploadIntervalHoursCache",
    "NexLastUploadTrackBlockHardIndex",
    "MainScreenOnOff",
    "PlayReport_AudioChannel_Stereo",
    "NexIsSaveDataUploadAvailableCache",
    "NexLastUploadSaveDataTime",
    "PlayReport_AllPlayTime",
    "NexLastUploadTrackBlockIndex",
    "PlayReport_CtrlMode_Console_Ext",
    "AmiiboItemOnOff",
    "TrackBlockFileNumber_Hard",
    "StickSensitivity",
    "TextWindowChange",
    "IsLastPlayHardMode",
    "PlayReport_CtrlMode_Console_FullKey",
    "NexLastUploadTrackBlockTime",
    "PlayReport_CtrlMode_FullKey",
    "PlayReport_PlayTime_Console",
    "PlayReport_AudioChannel_Mono",
    "CameraUpDownReverse",
    "PlayReport_CtrlMode_Handheld",
]


class FlagStore:
    _store: Dict[str, Dict[int, BFUFlag]]
    _orig_store: Dict[str, Dict[int, BFUFlag]]

    def __init__(self) -> None:
        self._store = {}
        self._orig_store = {}
        for ftype in FLAG_MAPPING:
            self._store[ftype] = {}
            self._orig_store[ftype] = {}

    def add_flags_from_Hash(self, name: str, data: Hash) -> None:
        is_revival = bool("revival" in name)
        for ftype, flags in data.items():
            for flag in flags:
                self._store[ftype][flag["HashValue"].v] = FLAG_MAPPING[ftype](
                    flag, revival=is_revival
                )
                self._orig_store[ftype][flag["HashValue"].v] = FLAG_MAPPING[ftype](
                    flag, revival=is_revival
                )

    def add_flags_from_Hash_no_overwrite(self, name: str, data: Hash) -> None:
        is_revival = bool("revival" in name)
        for ftype, flags in data.items():
            for flag in flags:
                if not self.find(ftype, flag["HashValue"].v).exists():
                    self._store[ftype][flag["HashValue"].v] = FLAG_MAPPING[ftype](
                        flag, revival=is_revival
                    )
                    self._orig_store[ftype][flag["HashValue"].v] = FLAG_MAPPING[ftype](
                        flag, revival=is_revival
                    )

    def find(self, ftype: str, hash: int) -> BFUFlag:
        if hash in self._store[ftype]:
            return self._store[ftype][hash]
        return BFUFlag()

    def find_all(self, ftype: str, search: str) -> List[BFUFlag]:
        r: List[BFUFlag] = []
        for _, flag in self._store[ftype].items():
            if flag.name_contains(search):
                r.append(flag)
        return r

    def find_all_hashes(self, ftype: str, search: str) -> Set[int]:
        r: Set[int] = set()
        for hash, flag in self._store[ftype].items():
            if flag.name_contains(search):
                r.add(hash)
        return r

    def add(self, ftype: str, flag: BFUFlag) -> None:
        self._store[ftype][flag.hash_value] = flag

    def remove(self, ftype: str, hash: int) -> None:
        self._store[ftype].pop(hash, None)

    def get_num_new(self) -> int:
        r = 0
        for ftype in FLAG_MAPPING:
            r += len(self.get_new_ftype(ftype))
        return r

    def get_num_modified(self) -> int:
        r = 0
        for ftype in FLAG_MAPPING:
            r += len(self.get_modified_ftype(ftype))
        return r

    def get_num_deleted(self) -> int:
        r = 0
        for ftype in FLAG_MAPPING:
            r += len(self.get_deleted_ftype(ftype))
        return r

    def get_new_ftype(self, ftype: str) -> Set[str]:
        r: Set[str] = set()
        for _, flag in self._store[ftype].items():
            if flag.hash_value not in self._orig_store[ftype]:
                r.add(flag.data_name)
        return r

    def get_modified_ftype(self, ftype: str) -> Set[str]:
        r: Set[str] = set()
        for _, flag in self._store[ftype].items():
            if flag.hash_value in self._orig_store[ftype]:
                if not flag == self._orig_store[ftype][flag.hash_value]:
                    r.add(flag.data_name)
        return r

    def get_deleted_ftype(self, ftype: str) -> Set[str]:
        r: Set[str] = set()
        for _, flag in self._orig_store[ftype].items():
            if flag.hash_value not in self._store[ftype]:
                r.add(flag.data_name)
        return r

    def get_total_changes(self) -> int:
        return self.get_num_new() + self.get_num_modified() + self.get_num_deleted()

    def get_num_new_svdata(self) -> int:
        r = 0
        for ftype in FLAG_MAPPING:
            r += len(self.get_new_ftype_svdata(ftype))
        return r

    def get_num_modified_svdata(self) -> int:
        """Always returns 0 because modifying anything about
        a save flag will make it a new+delete instead"""
        return 0

    def get_num_deleted_svdata(self) -> int:
        r = 0
        for ftype in FLAG_MAPPING:
            r += len(self.get_deleted_ftype_svdata(ftype))
        return r

    def get_new_ftype_svdata(self, ftype: str) -> Set[str]:
        r: Set[str] = set()
        for _, flag in self._store[ftype].items():
            if flag.is_save:
                if flag.hash_value not in self._orig_store[ftype]:
                    r.add(flag.data_name)
        return r

    def get_modified_ftype_svdata(self, ftype: str) -> Set[str]:
        """Always returns empty because modifying anything about
        a save flag will make it a new+delete instead"""
        return set()

    def get_deleted_ftype_svdata(self, ftype: str) -> Set[str]:
        r: Set[str] = set()
        for _, flag in self._orig_store[ftype].items():
            if flag.is_save:
                if flag.hash_value not in self._store[ftype]:
                    r.add(flag.data_name)
        return r

    def flags_to_bgdata_Array(self, prefix: str) -> Array:
        ftype = BGDATA_MAPPING[prefix]
        if prefix == "revival_bool_data" or prefix == "revival_s32_data":
            flag_list = [
                flag.to_Hash() for _, flag in self._store[ftype].items() if flag.is_revival
            ]
        elif prefix == "bool_data" or prefix == "s32_data":
            flag_list = [
                flag.to_Hash() for _, flag in self._store[ftype].items() if not flag.is_revival
            ]
        else:
            flag_list = [flag.to_Hash() for _, flag in self._store[ftype].items()]

        return Array(sorted(flag_list, key=lambda f: f["HashValue"]))

    def flags_to_svdata_Array(self) -> Array:
        flag_list: list = []
        for _, flagdict in self._store.items():
            flag_list += [
                flag.to_sv_Hash()
                for _, flag in flagdict.items()
                if flag.is_save and not flag.data_name in IGNORED_SAVE_FLAGS
            ]
        return Array(sorted(flag_list, key=lambda f: f["HashValue"]))



from oead.yaz0 import compress
from pathlib import Path

from . import util
from .flag import BFUFlag
from .store import FlagStore


def find(args):
    util.root_dir(args.directory)
    gamedata_sarc = util.get_gamedata_sarc()
    bgdata = FlagStore()
    for bgdata_name, bgdata_hash in map(util.unpack_oead_file, gamedata_sarc.get_files()):
        bgdata.add_flags_from_Hash(bgdata_name, bgdata_hash)

    search_name = args.flag_name
    found: dict = {}
    numfound = 0
    numsv = 0
    for ftype in util.BGDATA_TYPES:
        found[ftype] = bgdata.find_all(ftype, search_name)
        numfound += len(found[ftype])
    for _, flags in found.items():
        for flag in flags:
            if flag.is_save:
                numsv += 1

    while True:
        print(
            f"\n{numfound} gamedata flags and {numsv} savedata flags were found that matched {search_name}."
        )
        print("\nPlease choose an option:")
        print("v - View the full flag names, files, and indices in their files")
        print("d - Delete these flags and exit")
        print("x - Exit without deleting the flags")
        selection = input("(v/d/x):")

        if selection == "v":
            for ftype, flags in found.items():
                for flag in flags:
                    if flag.is_save:
                        string = f"{flag.data_name} in {ftype} and in game_data.sav"
                    else:
                        string = f"{flag.data_name} in {ftype}"
                    print(string)

        elif selection == "d":
            for ftype, flags in found.items():
                for flag in flags:
                    bgdata.remove(ftype, flag.hash_value)
            files_to_write: list = []
            files_to_write.append("GameData/gamedata.ssarc")
            files_to_write.append("GameData/savedataformat.ssarc")
            orig_files = util.get_last_two_savedata_files()
            datas_to_write: list = []
            datas_to_write.append(compress(util.make_new_gamedata(bgdata, args.bigendian)))
            datas_to_write.append(
                compress(util.make_new_savedata(bgdata, args.bigendian, orig_files))
            )
            util.inject_files_into_bootup(files_to_write, datas_to_write)
            return

        elif selection == "x":
            return



from math import ceil, sqrt
from pathlib import Path
from typing import Union
from time import time

import oead
from bcml import util as bcmlutil
from . import BGDATA_MAPPING, vanilla_shrine_locs
from .store import FlagStore


BGDATA_TYPES = [
    "bool_array_data",
    "bool_data",
    "f32_array_data",
    "f32_data",
    "s32_array_data",
    "s32_data",
    "string256_array_data",
    "string256_data",
    "string_data",
    "string64_array_data",
    "string64_data",
    "vector2f_array_data",
    "vector2f_data",
    "vector3f_array_data",
    "vector3f_data",
    "vector4f_data",
]


def root_dir(dir: str = "") -> Path:
    if not dir == "":
        root_dir._root_dir = dir  # type:ignore[attr-defined]
    if not hasattr(root_dir, "_root_dir"):
        raise RuntimeError("Root directory was never set.")
    return Path(root_dir._root_dir)  # type:ignore[attr-defined]


def convert_to_vec3f(hash: Union[dict, oead.byml.Hash]) -> oead.Vector3f:
    vec = oead.Vector3f()
    if type(hash) == dict:
        vec.x = hash["X"]
        vec.y = hash["Y"]
        vec.z = hash["Z"]
    elif type(hash) == oead.byml.Hash:
        vec.x = hash["X"].v
        vec.y = hash["Y"].v
        vec.z = hash["Z"].v
    return vec


def get_vector_distance(vec1: oead.Vector3f, vec2: oead.Vector3f) -> float:
    return sqrt((vec1.x - vec2.x) ** 2 + (vec1.y - vec2.y) ** 2 + (vec1.z - vec2.z) ** 2)


def get_shrine_locs() -> dict:
    if not hasattr(get_shrine_locs, "_shrine_locs"):
        shrine_locs = {
            shrine: convert_to_vec3f(loc) for shrine, loc in vanilla_shrine_locs.items()
        }
        static_path = root_dir() / "aoc/0010/Map/MainField/Static.smubin"
        if static_path.exists():
            static = oead.byml.from_binary(oead.yaz0.decompress(static_path.read_bytes()))
            for marker in static["LocationMarker"]:
                if not "Icon" in marker:
                    continue
                if not marker["Icon"] == "Dungeon":
                    continue
                if "MessageID" in marker:
                    if not marker["MessageID"] in vanilla_shrine_locs:
                        shrine_locs[marker["MessageID"]] = convert_to_vec3f(marker["Translate"])
        get_shrine_locs._shrine_locs = shrine_locs  # type:ignore[attr-defined]
    return get_shrine_locs._shrine_locs  # type:ignore[attr-defined]


def get_nearest_shrine(vec: oead.Vector3f) -> str:
    shrine_locs = get_shrine_locs()
    smallest_distance: float = 10000000.0
    nearest_shrine: str = ""
    for name, loc in shrine_locs.items():
        shrine_distance = get_vector_distance(vec, loc)
        if shrine_distance < smallest_distance:
            nearest_shrine = name
            smallest_distance = shrine_distance
    return nearest_shrine


def get_gamedata_sarc() -> oead.Sarc:
    bootup_path: Path = root_dir() / "content" / "Pack" / "Bootup.pack"
    bootup_sarc = oead.Sarc(bootup_path.read_bytes())
    gamedata_sarc = oead.Sarc(
        oead.yaz0.decompress(bootup_sarc.get_file("GameData/gamedata.ssarc").data)
    )
    return gamedata_sarc


def get_last_two_savedata_files() -> list:
    bootup_path: Path = root_dir() / "content" / "Pack" / "Bootup.pack"
    bootup_sarc = oead.Sarc(bootup_path.read_bytes())
    savedata_sarc = oead.Sarc(
        oead.yaz0.decompress(bootup_sarc.get_file("GameData/savedataformat.ssarc").data)
    )
    savedata_writer = oead.SarcWriter.from_sarc(savedata_sarc)
    idx = 0
    files = []
    while True:
        try:
            savedata_writer.files[f"/saveformat_{idx+2}.bgsvdata"]
            idx += 1
        except KeyError:
            files.append(savedata_writer.files[f"/saveformat_{idx}.bgsvdata"])
            files.append(savedata_writer.files[f"/saveformat_{idx+1}.bgsvdata"])
            return files


def make_new_gamedata(store: FlagStore, big_endian: bool) -> bytes:
    bgwriter = oead.SarcWriter(
        endian=oead.Endianness.Big if big_endian else oead.Endianness.Little
    )
    for prefix, data_type in BGDATA_MAPPING.items():
        bgdata_array = store.flags_to_bgdata_Array(prefix)
        num_files = ceil(len(bgdata_array) / 4096)
        for idx in range(num_files):
            start = idx * 4096
            end = (idx + 1) * 4096
            if end > len(bgdata_array):
                end = len(bgdata_array)
            bgwriter.files[f"/{prefix}_{idx}.bgdata"] = oead.byml.to_binary(
                oead.byml.Hash({data_type: bgdata_array[start:end]}), big_endian,
            )
    return bgwriter.write()[1]


def make_new_savedata(store: FlagStore, big_endian: bool, orig_files: list) -> bytes:
    svwriter = oead.SarcWriter(
        endian=oead.Endianness.Big if big_endian else oead.Endianness.Little
    )
    svdata_array = store.flags_to_svdata_Array()
    num_files = ceil(len(svdata_array) / 8192)
    for idx in range(num_files):
        start = idx * 8192
        end = (idx + 1) * 8192
        if end > len(svdata_array):
            end = len(svdata_array)
        svwriter.files[f"/saveformat_{idx}.bgsvdata"] = oead.byml.to_binary(
            oead.byml.Hash(
                {
                    "file_list": oead.byml.Array(
                        [
                            {
                                "IsCommon": False,
                                "IsCommonAtSameAccount": False,
                                "IsSaveSecureCode": True,
                                "file_name": "game_data.sav",
                            },
                            oead.byml.Array(svdata_array[start:end]),
                        ]
                    ),
                    "save_info": oead.byml.Array(
                        [
                            {
                                "directory_num": oead.S32(num_files + 2),
                                "is_build_machine": True,
                                "revision": oead.S32(18203),
                            }
                        ]
                    ),
                }
            ),
            big_endian,
        )
    svwriter.files[f"/saveformat_{num_files}.bgsvdata"] = orig_files[0]
    svwriter.files[f"/saveformat_{num_files+1}.bgsvdata"] = orig_files[1]
    return svwriter.write()[1]


def inject_files_into_bootup(files: list, datas: list):
    bootup_path: Path = root_dir() / "content" / "Pack" / "Bootup.pack"
    sarc_data = bootup_path.read_bytes()
    yaz = sarc_data[0:4] == b"Yaz0"
    if yaz:
        sarc_data = bcmlutil.decompress(sarc_data)
    old_sarc = oead.Sarc(sarc_data)
    del sarc_data
    new_sarc = oead.SarcWriter.from_sarc(old_sarc)
    del old_sarc
    for idx in range(len(files)):
        new_sarc.files[files[idx]] = (
            datas[idx] if isinstance(datas[idx], bytes) else bytes(datas[idx])
        )
    new_bytes = new_sarc.write()[1]
    del new_sarc
    bootup_path.write_bytes(new_bytes if not yaz else bcmlutil.compress(new_bytes))
    del new_bytes


def unpack_oead_file(f: oead.File) -> tuple:
    return (f.name, oead.byml.from_binary(f.data))


def get_verbose_output(store: FlagStore) -> str:
    r: list = []
    for ftype in BGDATA_TYPES:
        r.append(f"For {ftype}:\n")
        r.append("  Game data entries:\n")
        r.append("    New flags:\n")
        for name in sorted(list(store.get_new_ftype(ftype))):
            r.append(f"      {name}\n")
        r.append("    Modified flags:\n")
        for name in sorted(list(store.get_modified_ftype(ftype))):
            r.append(f"      {name}\n")
        r.append("    Deleted flags:\n")
        for name in sorted(list(store.get_deleted_ftype(ftype))):
            r.append(f"      {name}\n")
        r.append("  Save data entries:\n")
        r.append("    New flags:\n")
        for name in sorted(list(store.get_new_ftype_svdata(ftype))):
            r.append(f"      {name}\n")
        r.append("    Modified flags:\n")
        for name in sorted(list(store.get_modified_ftype_svdata(ftype))):
            r.append(f"      {name}\n")
        r.append("    Deleted flags:\n")
        for name in sorted(list(store.get_deleted_ftype_svdata(ftype))):
            r.append(f"      {name}\n")
    return "".join(r)



import ctypes
import json
import time
import zlib
from math import ceil
from pathlib import Path

import oead
from bcml.mergers import mubin
from bcml import util as bcmlutil

from . import (
    EXEC_DIR,
    BGDATA_MAPPING,
    util,
    vanilla_hash_dict,
    vanilla_actors,
    vanilla_shrine_locs,
)
from .flag import BoolFlag, S32Flag
from .store import FlagStore


current_map: str = ""
bgdata: FlagStore = FlagStore()
mod_actors_with_life: set = set()
added_flag_hashes: set = set()
orphaned_flag_hashes: set = set()
GENERATOR_FLAG_NAME_EXCEPTIONS: list = [
    "DgnMrgPrt",
]


def should_make_flag(obj) -> bool:
    if "!Parameters" in obj:
        if "ForceFlag" in obj["!Parameters"]:
            return obj["!Parameters"]["ForceFlag"]

    if "LinkTag" in obj["UnitConfigName"]:
        try:
            makeflag = obj["!Parameters"]["MakeSaveFlag"].v
        except KeyError:
            return False
        if makeflag == 0 and not "SaveFlag" in obj["!Parameters"]:
            return False
        return True

    for substr in GENERATOR_FLAG_NAME_EXCEPTIONS:
        if substr in obj["UnitConfigName"]:
            return False

    if obj["UnitConfigName"] in mod_actors_with_life:
        return True

    if obj["UnitConfigName"] in vanilla_actors["with_flags"]:
        return True
    return False


def get_flag_name(obj, maptype: str) -> str:
    if "LinkTag" in obj["UnitConfigName"] and "!Parameters" in obj:
        try:
            makeflag = obj["!Parameters"]["MakeSaveFlag"].v
        except KeyError:
            return ""
        if makeflag == 0 and "SaveFlag" in obj["!Parameters"]:
            return obj["!Parameters"]["SaveFlag"]
            # should_make_flag() will stop cases where `not "SaveFlag" in obj["!Parameters"]`
        elif makeflag == 1:
            if not current_map:
                raise ValueError(
                    "A LinkTag was created with MakeSaveFlag 1 in MainField, this is not valid"
                )
            return f"Clear_{current_map}"
        elif makeflag == 2:
            if current_map:
                raise ValueError(
                    "A LinkTag was created with MakeSaveFlag 2 in CDungeon, this is not valid"
                )
            loc = oead.Vector3f()
            loc.x = obj["Translate"][0].v
            loc.y = obj["Translate"][1].v
            loc.z = obj["Translate"][2].v
            return f"Open_{util.get_nearest_shrine(loc)}"
    return f"{maptype}_{obj['UnitConfigName']}_{obj['HashId'].v}"


def bool_flag(new_obj, old_obj, maptype: str, resettype: int = 1, revival: bool = True) -> None:
    old_hash: int = ctypes.c_int32(zlib.crc32(get_flag_name(old_obj, maptype).encode())).value
    new_name: str = get_flag_name(new_obj, maptype)

    if not should_make_flag(new_obj):
        bgdata.remove("bool_data", old_hash)
        new_hash = ctypes.c_int32(zlib.crc32(new_name.encode())).value
        bgdata.remove("bool_data", new_hash)
        return

    if "TBox" in new_obj["UnitConfigName"]:
        if "EnableRevival" in new_obj["!Parameters"]:
            resettype = int(new_obj["!Parameters"]["EnableRevival"])

    flag = bgdata.find("bool_data", old_hash)
    old_exists = flag.exists()
    if not old_exists:
        flag = bgdata.find("bool_data", ctypes.c_int32(zlib.crc32(new_name.encode())).value)
        if not flag.exists():
            flag = BoolFlag(revival=revival)
            flag.is_save = True
            flag.reset_type = resettype
    flag.data_name = new_name
    flag.use_name_to_override_params()

    if old_exists:
        bgdata.remove("bool_data", old_hash)
    bgdata.add("bool_data", flag)

    added_flag_hashes.add(flag.hash_value)


def s32_flag(new_obj, old_obj, maptype: str, resettype: int = 1, revival: bool = True) -> None:
    old_hash: int = ctypes.c_int32(zlib.crc32(get_flag_name(old_obj, maptype).encode())).value
    new_name: str = get_flag_name(new_obj, maptype)

    if not should_make_flag(new_obj):
        bgdata.remove("s32_data", old_hash)
        new_hash = ctypes.c_int32(zlib.crc32(new_name.encode())).value
        bgdata.remove("s32_data", new_hash)
        return

    flag = bgdata.find("s32_data", old_hash)
    old_exists = flag.exists()
    if not old_exists:
        flag = bgdata.find("s32_data", ctypes.c_int32(zlib.crc32(new_name.encode())).value)
        if not flag.exists():
            flag = S32Flag(revival=revival)
            flag.is_save = True
            flag.reset_type = resettype
    flag.data_name = new_name
    flag.use_name_to_override_params()

    if old_exists:
        bgdata.remove("s32_data", old_hash)
    bgdata.add("s32_data", flag)

    added_flag_hashes.add(flag.hash_value)


def misc_s32_flag(name: str) -> None:
    flag = S32Flag()
    flag.data_name = name
    flag.use_name_to_override_params()

    bgdata.add("s32_data", flag)

    added_flag_hashes.add(flag.hash_value)


def misc_bool_flag(name: str) -> None:
    flag = BoolFlag()
    flag.data_name = name
    flag.use_name_to_override_params()

    bgdata.add("bool_data", flag)

    added_flag_hashes.add(flag.hash_value)


def generate_revival_flags_for_map(
    map_data: oead.byml.Hash, stock_map: oead.byml.Hash, maptype: str, resettype: int
) -> None:

    map_hashes = [obj["HashId"].v for obj in map_data["Objs"]]
    stock_hashes = [obj["HashId"].v for obj in stock_map["Objs"]]
    for obj in map_data["Objs"]:
        revival = True
        bflag = True
        if "LinkTag" in obj["UnitConfigName"]:
            revival = False
            if not "!Parameters" in obj:
                continue  # if it's a LinkTag with no !Parameters, fuck that developer
            if "IncrementSave" in obj["!Parameters"]:
                if obj["!Parameters"]["IncrementSave"]:
                    bflag = False
        if obj["HashId"].v not in stock_hashes:
            if bflag:
                bool_flag(obj, obj, maptype, resettype, revival)
            else:
                s32_flag(obj, obj, maptype, resettype, revival)
        elif obj["HashId"].v in stock_hashes:
            stock_obj = stock_map["Objs"][stock_hashes.index(obj["HashId"].v)]
            if bflag:
                bool_flag(obj, stock_obj, maptype, resettype, revival)
            else:
                s32_flag(obj, stock_obj, maptype, resettype, revival)
    for obj in stock_map["Objs"]:
        if obj["HashId"].v not in map_hashes:
            old_hash: int = ctypes.c_int32(
                zlib.crc32(f"{maptype}_{obj['UnitConfigName']}_{obj['HashId'].v}".encode())
            ).value
            orphaned_flag_hashes.add(old_hash)


def generate_revival_flags(resettypes: list) -> None:
    moddir: Path = util.root_dir()

    if not resettypes[0] == -1:
        for map_unit in moddir.rglob("*_*.smubin"):
            map_start = time.time()
            map_data = oead.byml.from_binary(oead.yaz0.decompress(map_unit.read_bytes()))
            map_section = map_unit.stem.split("_")
            stock_map = mubin.get_stock_map((map_section[0], map_section[1]))
            generate_revival_flags_for_map(map_data, stock_map, "MainField", resettypes[0])
            print(f"Finished processing {map_unit.name} in {time.time() - map_start} seconds...")
        for static_unit in moddir.rglob("MainField/Static.smubin"):
            map_start = time.time()
            static_data = oead.byml.from_binary(oead.yaz0.decompress(static_unit.read_bytes()))
            for marker in static_data["LocationMarker"]:
                if not "Icon" in marker:
                    continue
                if not marker["Icon"] == "Dungeon":
                    continue
                if "MessageID" in marker:
                    if not marker["MessageID"] in vanilla_shrine_locs:
                        misc_s32_flag(marker["SaveFlag"])
                        misc_bool_flag(f"Enter_{marker['MessageID']}")
                        misc_bool_flag(f"CompleteTreasure_{marker['MessageID']}")
            print(
                f"Finished processing MainField/Static.smubin in {time.time() - map_start} seconds..."
            )
    if not resettypes[1] == -1:
        global current_map
        for map_pack in moddir.rglob("Pack/Dungeon*.pack"):
            current_map = map_pack.stem
            pack_data = oead.Sarc(map_pack.read_bytes())
            try:
                stock_pack = oead.Sarc(bcmlutil.get_game_file(map_pack).read_bytes())
            except FileNotFoundError:
                stock_pack = None
            map_types = ("_Static", "_Dynamic")
            for map_type in map_types:
                map_start = time.time()
                map_name = f"{map_pack.stem}{map_type}.smubin"
                map_data = oead.byml.from_binary(
                    oead.yaz0.decompress(
                        pack_data.get_file(f"Map/CDungeon/{map_pack.stem}/{map_name}").data
                    )
                )
                if stock_pack:
                    stock_map = oead.byml.from_binary(
                        oead.yaz0.decompress(
                            stock_pack.get_file(f"Map/CDungeon/{map_pack.stem}/{map_name}").data
                        )
                    )
                else:
                    stock_map = oead.byml.Hash()
                    stock_map["Objs"] = oead.byml.Array()
                generate_revival_flags_for_map(map_data, stock_map, "CDungeon", resettypes[1])
                print(f"Finished processing {map_name} in {time.time() - map_start} seconds...")
            current_map = ""
            # ^ unnecessary given that it's updated every loop, but we rely on it possibly being
            # empty for determining whether LinkTags are valid, so it's better to be safe
    for hash in orphaned_flag_hashes:
        if hash not in added_flag_hashes:
            bgdata.remove("bool_data", hash)


def actor_bool_flag(flag_name: str) -> int:
    flag = BoolFlag()
    flag.data_name = flag_name
    flag.use_name_to_override_params()

    bgdata.add("bool_data", flag)

    return flag.hash_value


def actor_s32_flag(flag_name: str) -> int:
    flag = S32Flag()
    flag.data_name = flag_name
    flag.use_name_to_override_params()

    bgdata.add("s32_data", flag)

    return flag.hash_value


def generate_item_flags() -> None:
    moddir: Path = util.root_dir()
    mod_bool: set = set()
    mod_s32: set = set()
    for actor in moddir.rglob("*.sbactorpack"):
        if "Animal_" in actor.name:
            mod_bool.add(actor_bool_flag(f"IsNewPictureBook_{actor.stem}"))
            mod_bool.add(actor_bool_flag(f"IsRegisteredPictureBook_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"PictureBookSize_{actor.stem}"))
            bxml = oead.aamp.ParameterIO.from_binary(
                oead.Sarc(oead.yaz0.decompress(actor.read_bytes()))
                .get_file(f"Actor/ActorLink/{actor.stem}.bxml")
                .data
            )
            if "Tags" in bxml.objects:
                for _, tag in bxml.objects["Tags"].params.items():
                    if tag == "CanGetPouch":
                        mod_bool.add(actor_bool_flag(f"IsGet_{actor.stem}"))
                        break
            del bxml
        elif "Armor_" in actor.name:
            mod_bool.add(actor_bool_flag(f"IsGet_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"EquipTime_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"PorchTime_{actor.stem}"))
        elif "Enemy_" in actor.name:
            mod_bool.add(actor_bool_flag(f"IsNewPictureBook_{actor.stem}"))
            mod_bool.add(actor_bool_flag(f"IsRegisteredPictureBook_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"PictureBookSize_{actor.stem}"))
        elif "Item_" in actor.name:
            mod_bool.add(actor_bool_flag(f"IsNewPictureBook_{actor.stem}"))
            mod_bool.add(actor_bool_flag(f"IsRegisteredPictureBook_{actor.stem}"))
            mod_bool.add(actor_bool_flag(f"IsGet_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"PictureBookSize_{actor.stem}"))
        elif "Npc_" in actor.name:
            sarc = oead.Sarc(oead.yaz0.decompress(actor.read_bytes()))
            bxml = oead.aamp.ParameterIO.from_binary(
                sarc.get_file(f"Actor/ActorLink/{actor.stem}.bxml").data
            )
            shop_link = bxml.objects["LinkTarget"].params["ShopDataUser"]
            del bxml
            if shop_link == "Dummy":
                del sarc
                continue
            mod_bool.add(actor_bool_flag(f"{actor.stem}_SoldOut"))
            bshop = oead.aamp.ParameterIO.from_binary(
                sarc.get_file(f"Actor/ShopData/{shop_link}.bshop").data
            )
            del sarc
            for tablekey, tablename in bshop.objects["Header"].params.items():
                if tablekey.hash == zlib.crc32(b"TableNum"):
                    continue
                for _, val in bshop.objects[str(tablename)].params.items():
                    if isinstance(val, oead.FixedSafeString64):
                        mod_s32.add(actor_s32_flag(f"{actor.stem}_{str(val)}"))
            del bshop
        elif "Weapon_" in actor.name:
            mod_bool.add(actor_bool_flag(f"IsNewPictureBook_{actor.stem}"))
            mod_bool.add(actor_bool_flag(f"IsRegisteredPictureBook_{actor.stem}"))
            mod_bool.add(actor_bool_flag(f"IsGet_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"PictureBookSize_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"EquipTime_{actor.stem}"))
            mod_s32.add(actor_s32_flag(f"PorchTime_{actor.stem}"))

    """
    This code section is meant to delete flags for actors that have been removed.
    However, the likelihood of this ever actually needing to happen is so low,
    and the cost of maintaining it so comparatively high, that it's being removed
    for now. I might explore reimplementing it at a later date.

    vanilla_hashes: set = set()
    for _, hash_list in vanilla_hash_dict.items():
        vanilla_hashes |= set(hash_list)

    total_bool: set = set()
    total_bool |= bgdata.find_all_hashes("bool_data", "IsNewPictureBook_")
    total_bool |= bgdata.find_all_hashes("bool_data", "IsRegisteredPictureBook_")
    total_bool |= bgdata.find_all_hashes("bool_data", "IsGet_")
    to_delete = total_bool - (mod_bool | vanilla_hashes)
    for hash in to_delete:
        bgdata.remove("bool_data", hash)

    total_s32: set = set()
    total_s32 |= bgdata.find_all_hashes("s32_data", "PictureBookSize_")
    total_s32 |= bgdata.find_all_hashes("s32_data", "EquipTime_")
    total_s32 |= bgdata.find_all_hashes("s32_data", "PorchTime_")
    to_delete = total_s32 - (mod_s32 | vanilla_hashes)
    for hash in to_delete:
        bgdata.remove("s32_data", hash)
    """


def generate(args):
    if not args.actor and args.revival[0] == -1 and args.revival[1] == -1:
        print("No flag options were chosen! Use -a and/or -r to generate flags.")
        exit()

    util.root_dir(args.directory)
    actorinfo_path = util.root_dir() / "content/Actor/ActorInfo.product.sbyml"
    if actorinfo_path.exists():
        actorinfo_bytes = actorinfo_path.read_bytes()
        actorinfo = oead.byml.from_binary(oead.yaz0.decompress(actorinfo_bytes))
        for actor in actorinfo["Actors"]:
            if (
                not actor["name"] in vanilla_actors["with_flags"]
                and not actor["name"] in vanilla_actors["no_flags"]
                and "generalLife" in actor
            ):
                mod_actors_with_life.add(actor["name"])
        del actorinfo_bytes
        del actorinfo

    gamedata_sarc = util.get_gamedata_sarc()
    for bgdata_name, bgdata_hash in map(util.unpack_oead_file, gamedata_sarc.get_files()):
        bgdata.add_flags_from_Hash(bgdata_name, bgdata_hash)

    if args.revival:
        generate_revival_flags(args.revival)
    if args.actor:
        generate_item_flags()

    files_to_write: list = []
    files_to_write.append("GameData/gamedata.ssarc")
    files_to_write.append("GameData/savedataformat.ssarc")
    orig_files = util.get_last_two_savedata_files()
    datas_to_write: list = []
    bgdata_start = time.time()
    datas_to_write.append(oead.yaz0.compress(util.make_new_gamedata(bgdata, args.bigendian)))
    bgdata_time = time.time() - bgdata_start
    print(f"Generating bgdata took {bgdata_time} seconds...")
    bgdata_start = time.time()
    datas_to_write.append(
        oead.yaz0.compress(util.make_new_savedata(bgdata, args.bigendian, orig_files))
    )
    bgdata_time = time.time() - bgdata_start
    print(f"Generating svdata took {bgdata_time} seconds...")
    util.inject_files_into_bootup(files_to_write, datas_to_write)

    if bgdata.get_total_changes() > 0:
        print()
        print(f"{bgdata.get_num_new()} New Game Data Entries")
        print(f"{bgdata.get_num_modified()} Modified Game Data Entries")
        print(f"{bgdata.get_num_deleted()} Deleted Game Data Entries")
        print(f"{bgdata.get_num_new_svdata()} New Save Data Entries")
        print(f"{bgdata.get_num_modified_svdata()} Modified Save Data Entries")
        print(f"{bgdata.get_num_deleted_svdata()} Deleted Save Data Entries")

        if args.verbose:
            flag_path = util.root_dir() / "flag_log.txt"
            flag_path.touch()
            flag_path.write_text(util.get_verbose_output(bgdata))
            flag_dir = str(flag_path).replace("\\", "/")
            print()
            print(f"Log written to {flag_dir}")
    else:
        print("No changes were made.")
